import time
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Union

from .prefix_metadata import PrefixMetadata, PrefixMetadataDataSource


@dataclass()
class QueueItemStats:
    prefix_first: str
    prefix_last: str
    prefix_count: int

    request_count: int
    bytes_received: int

    local_source_ttl_cache_count: int
    local_source_etag_match_count: int
    remote_source_remote_cache_count: int
    remote_source_origin_source_count: int
    unknown_source_status_count: int

    start_time: float = field(default=time.time())
    __end_time: Union[float, None] = field(default=None)
    __request_rate: float = field(default=0)  # per-second
    __bytes_received_rate: float = field(default=0)  # per-second

    @property
    def request_rate(self):
        return self.__request_rate

    @property
    def bytes_received_rate(self):
        return self.__bytes_received_rate

    @property
    def run_time(self):
        return self.end_time - self.start_time

    @property
    def end_time(self):
        return self.__end_time

    @end_time.setter
    def end_time(self, value: float):
        self.__end_time = value
        if self.run_time > 0:
            self.__request_rate = self.request_count / self.run_time
            self.__bytes_received_rate = self.bytes_received / self.run_time

    def end_trigger(self):
        self.end_time = time.time()


@dataclass()
class QueueRunningStats:
    start_time: float = field(default=time.time())
    queue_item_count = 0

    prefix_latest: Union[str, None] = field(default=None)
    prefix_count_sum = 0

    request_count_sum = 0
    bytes_received_sum = 0

    local_source_ttl_cache_count_sum = 0
    local_source_etag_match_count_sum = 0
    remote_source_remote_cache_count_sum = 0
    remote_source_origin_source_count_sum = 0
    unknown_source_status_count_sum = 0

    __end_time: Union[float, None] = field(default=None)
    __request_rate_total: float = field(default=0)  # per-second
    __bytes_received_rate_total: float = field(default=0)  # per-second

    @property
    def request_rate_total(self):
        return self.__request_rate_total

    @property
    def bytes_received_rate_total(self):
        return self.__bytes_received_rate_total

    @property
    def run_time(self):
        return self.end_time - self.start_time

    @property
    def end_time(self):
        return self.__end_time

    @end_time.setter
    def end_time(self, value: float):
        self.__end_time = value
        if self.run_time > 0:
            self.__request_rate_total = self.request_count_sum / self.run_time
            self.__bytes_received_rate_total = self.bytes_received_sum / self.run_time

    def end_trigger(self):
        self.end_time = time.time()

    def add_item_stats(self, item: QueueItemStats):
        self.queue_item_count += 1
        self.prefix_latest = item.prefix_last
        self.prefix_count_sum += item.request_count
        self.request_count_sum += item.request_count
        self.bytes_received_sum += item.bytes_received
        self.local_source_ttl_cache_count_sum += item.local_source_ttl_cache_count
        self.local_source_etag_match_count_sum += item.local_source_etag_match_count
        self.remote_source_remote_cache_count_sum += item.remote_source_remote_cache_count
        self.remote_source_origin_source_count_sum += item.remote_source_origin_source_count
        self.unknown_source_status_count_sum += item.unknown_source_status_count
        self.end_trigger()


class QueueItemStatsCompute:
    stats: QueueItemStats

    def __init__(self, results: List[PrefixMetadata]):
        data = {
            "start_time": None,
            "prefix_first": results[0].prefix,
            "prefix_last": results[-1].prefix,
            "prefix_count": len(results),
            "request_count": 0,
            "bytes_received": 0,
            "local_source_ttl_cache_count": 0,
            "local_source_etag_match_count": 0,
            "remote_source_remote_cache_count": 0,
            "remote_source_origin_source_count": 0,
            "unknown_source_status_count": 0,
        }

        oldest_timestamp = datetime.now().astimezone()

        for item in results:
            if item.start_timestamp and item.start_timestamp < oldest_timestamp:
                oldest_timestamp = item.start_timestamp

            if item.server_timestamp:
                data["request_count"] += 1

            if item.bytes and item.bytes > 0:
                data["bytes_received"] += item.bytes

            if item.data_source == PrefixMetadataDataSource.local_source_ttl_cache:
                data["local_source_ttl_cache_count"] += 1
            elif item.data_source == PrefixMetadataDataSource.local_source_etag_match:
                data["local_source_etag_match_count"] += 1
            elif item.data_source == PrefixMetadataDataSource.remote_source_remote_cache:
                data["remote_source_remote_cache_count"] += 1
            elif item.data_source == PrefixMetadataDataSource.remote_source_origin_source:
                data["remote_source_origin_source_count"] += 1
            elif item.data_source == PrefixMetadataDataSource.unknown_source_status:
                data["unknown_source_status_count"] += 1

        data["start_time"] = (oldest_timestamp - datetime.fromtimestamp(0).astimezone()).total_seconds()
        self.stats = QueueItemStats(**data)



#
# Copyright [2022-2023] Threat Patrols Pty Ltd (https://www.threatpatrols.com)
#

import atexit
from os import getenv
from signal import SIGINT, signal
from sys import argv

from typer import Exit as TyperExit

from . import __logger_name__, app_context
from .lib.app import invoke_app
from .lib.logger import logger_get

logger = logger_get(__logger_name__)
profiler = None  # type: ignore


def entrypoint():
    if "--profiler" in argv or getenv("HIBPDL_PROFILER", "").lower().startswith(("true", "yes", "enable")):
        logger.warning("pyinstrument.Profiler enabled")
        from pyinstrument import Profiler

        global profiler
        profiler = Profiler()
        profiler.start()
        app_context.profiler = True

    invoke_app()


def sigint_handler(__signal_received, __frame):
    print("SIGINT received, exiting.")
    raise TyperExit(code=1)


def exit_handler():
    if app_context.profiler:
        global profiler
        profiler.stop()
        output_filename = "profiler.html"
        output_html = profiler.output_html()
        with open(output_filename, "w") as f:
            f.write(output_html)
        logger.warning(f"pyinstrument.Profiler report written to {output_filename!r}")


signal(SIGINT, sigint_handler)
atexit.register(exit_handler)



from os import cpu_count, getenv
from sys import argv

from .lib.logger import logger_get
from .models import AppContext

__title__ = "HIBP Downloader"
__version__ = "0.1.4"
__logger_name__ = "hibp-downloader"
__pwnedpasswords_api_url__ = "https://api.pwnedpasswords.com"
__local_cache_ttl_default__ = 86400
__multiprocessing_processes_default__ = cpu_count()
__multiprocessing_prefixes_chunk_size__ = cpu_count() * 2
__approx_gzip_bytes_per_hash__ = 20.674
__logging_info_event_modulus__ = 5

# encoding_type
# The encoded response-content is stored as-is without trying to decode (ie decompress) it into a new encoding type
# for local storage; Because this content is stored as-is, it is more useful to use "gzip" because the command-line
# tools (eg zcat, zgrep) are more readily available than brotli enabled tools when examining the data-store files
__encoding_type__ = "gzip"  # values: [ gzip | br | None ]

__logger_level__ = "info"
if "--debug" in argv or getenv("HIBPDL_DEBUG", "").lower().startswith(("true", "yes", "enable")):
    __logger_level__ = "debug"

if "--quiet" in argv or getenv("HIBPDL_QUIET", "").lower().startswith(("true", "yes", "enable")):
    __logger_level__ = "fatal"

__app_profiler__ = False
if "--profiler" in argv or getenv("HIBPDL_PROFILER", "").lower().startswith(("true", "yes", "enable")):
    __app_profiler__ = True

__help_epilog_footer__ = f"""
{__title__} v{__version__}

Docs: [hibp-downloader.readthedocs.io](https://hibp-downloader.readthedocs.io)

Project: [github.com/threatpatrols/hibp-downloader](https://github.com/threatpatrols/hibp-downloader)
"""

logger_get(name=__logger_name__, loglevel=__logger_level__)
app_context = AppContext(debug=True if __logger_level__ == "debug" else False, profiler=__app_profiler__)



from dataclasses import dataclass, field
from typing import Union


@dataclass()
class AppContext:
    debug: bool = field(default=False)
    command: Union[str, None] = field(default=None)
    data_path: Union[str, None] = field(default=None)
    metadata_path: Union[str, None] = field(default=None)
    profiler: bool = False
    add_completion: bool = False
    no_args_is_help: bool = True



import asyncio
import os

import typer
from typing_extensions import Annotated

from hibp_downloader import (
    __encoding_type__,
    __help_epilog_footer__,
    __logger_name__,
    __logging_info_event_modulus__,
    app_context,
)
from hibp_downloader.exceptions import HibpDownloaderException
from hibp_downloader.lib.filedata import append_stringfile, load_datafile
from hibp_downloader.lib.generators import hex_sequence, iterable_chunker
from hibp_downloader.lib.logger import logger_get
from hibp_downloader.models import HashType

logger = logger_get(name=__logger_name__)

command = typer.Typer(no_args_is_help=False, epilog=__help_epilog_footer__)
command_name = "generate"
command_section = "Commands"


@command.callback(invoke_without_command=True)
def main(
    filename: Annotated[
        str,
        typer.Option(help="Filename to write the pwned password hash data as a single text file."),
    ],
    hash_type: Annotated[
        HashType,
        typer.Option(help="Hash type to use from the --data-path", case_sensitive=False),
    ] = HashType.sha1.value,
    first_hash: Annotated[
        str,
        typer.Option(help="Start the generator from a specific hash prefix; trimmed to the first 5 characters"),
    ] = "00000",
    last_hash: Annotated[
        str,
        typer.Option(help="Stop the generator at a hash prefix; trimmed to the first 5 characters"),
    ] = "fffff",
):
    """
    Generate a single text file with the pwned password hash values in-order from the --data-path location; [bold cyan]generate --help[/bold cyan] for more.
    """

    logger.debug(f"Starting command {app_context.command!r} from {os.path.basename(__file__)!r}")

    if not os.path.isdir(app_context.data_path):
        logger.error(f"Data path {app_context.data_path!r} does not exist, unable to continue")
        raise typer.Exit(1)

    if not os.path.isdir(app_context.metadata_path):
        logger.warning(f"Metadata path {app_context.metadata_path!r} does not exist, unable to continue")
        raise typer.Exit(1)

    logger.info(f"data-path {app_context.data_path!r}")

    if os.path.isfile(filename):
        logger.error(f"Output file already exists {filename!r}")
        raise typer.Exit(1)

    asyncio.run(pwnedpasswords_datastore_sorted_gather(filename, hash_type, first_hash[0:5], last_hash[0:5]))


async def pwnedpasswords_datastore_sorted_gather(filename, hash_type, first_hash, last_hash, chunk_size=16):
    iteration_count = 0
    for prefixes in iterable_chunker(iterable=hex_sequence(hex_first=first_hash, hex_last=last_hash), size=chunk_size):
        results = await asyncio.gather(
            *[pwnedpasswords_datastore_sorted_async(prefix, hash_type) for prefix in prefixes],
        )

        if (
            iteration_count == 0
            or iteration_count % (__logging_info_event_modulus__ * __logging_info_event_modulus__) == 0
        ):
            logger.info(f"Prefix position {prefixes[0]!r} appending to {filename!r}")
        iteration_count += 1

        output = ""
        for k, v in dict(sorted({k: v for d in results for k, v in d.items()}.items())).items():
            output = f"{output}\n{v}"
        await append_stringfile(filename=filename, content=output)


async def pwnedpasswords_datastore_sorted_async(prefix, hash_type):
    if __encoding_type__ in ("gz", "gzip"):
        filename_suffix = "gz"
        decompression_mode = "gzip"
    else:
        raise HibpDownloaderException(f"Unsupported __encoding_type__ {__encoding_type__}")

    source_data = await load_datafile(
        data_path=os.path.join(app_context.data_path, hash_type),
        prefix=prefix,
        filename_suffix=filename_suffix,
        decompression_type=decompression_mode,
        prepend_prefix=True,
    )

    return {prefix: source_data}



import asyncio
import json
import os

import typer
from typing_extensions import Annotated

from hibp_downloader import __encoding_type__, __help_epilog_footer__, __logger_name__, app_context
from hibp_downloader.exceptions import HibpDownloaderException
from hibp_downloader.lib.filedata import load_datafile
from hibp_downloader.lib.hashing import hashed_ntlm, hashed_sha1
from hibp_downloader.lib.logger import logger_get
from hibp_downloader.models import HashType

logger = logger_get(name=__logger_name__)

command = typer.Typer(no_args_is_help=False, epilog=__help_epilog_footer__)
command_name = "query"
command_section = "Commands"


@command.callback(invoke_without_command=True)
def main(
    password: Annotated[
        str,
        typer.Option(
            prompt=True,
            hide_input=True,
            help="Cleartext password string to query the local data store",
            envvar="HIBPDL_PASSWORD",
        ),
    ],
    hash_type: Annotated[
        HashType,
        typer.Option(help="Hash type to use from the --data-path", case_sensitive=False),
    ] = HashType.sha1.value,
):
    """
    Query the local --data-path and return any data that may exist; [bold cyan]query --help[/bold cyan] for more.
    """

    logger.debug(f"Starting command {app_context.command!r} from {os.path.basename(__file__)!r}")

    if not os.path.isdir(app_context.data_path):
        logger.error(f"Data path {app_context.data_path!r} does not exist, unable to continue")
        raise typer.Exit(1)

    if not os.path.isdir(app_context.metadata_path):
        logger.warning(f"Metadata path {app_context.metadata_path!r} does not exist, unable to continue")
        raise typer.Exit(1)

    logger.info(f"data-path {app_context.data_path!r}")

    if hash_type == HashType.sha1:
        password_hashed = hashed_sha1(password)
    elif hash_type == HashType.ntlm:
        password_hashed = hashed_ntlm(password)
    else:
        raise HibpDownloaderException(f"Unsupported hash_type {hash_type!r}")

    logger.debug(f"Password {hash_type.value!r} hashed {password_hashed!r}")

    asyncio.run(pwnedpasswords_query_datastore(password_hashed=password_hashed, hash_type=hash_type))


async def pwnedpasswords_query_datastore(password_hashed: str, hash_type: HashType):
    if __encoding_type__ in ("gz", "gzip"):
        filename_suffix = "gz"
        decompression_mode = "gzip"
    else:
        raise HibpDownloaderException(f"Unsupported __encoding_type__ {__encoding_type__}")

    prefix = password_hashed[0:5]

    result = {
        "hash": password_hashed.upper(),
        "hash_type": hash_type,
        "data_path": app_context.data_path,
        "hibp_count": None,
    }

    try:
        source_data = await load_datafile(
            data_path=os.path.join(app_context.data_path, hash_type),
            prefix=prefix,
            filename_suffix=filename_suffix,
            decompression_type=decompression_mode,
            prepend_prefix=True,
        )
    except HibpDownloaderException as e:
        result["status"] = str(e)
        return stdout_json(result)

    for line in source_data.split("\n"):
        if password_hashed.upper() in line:
            result["hibp_count"] = int(line.split(":")[1])
            result["status"] = "Found"
            return stdout_json(result)

    result["status"] = "Not Found"
    return stdout_json(result)


def stdout_json(data, indent="  "):
    as_string = json.dumps(data, indent=indent, sort_keys=True)
    print(as_string)
    return as_string



import httpx

from hibp_downloader import __logger_name__, __title__, __version__
from hibp_downloader.exceptions import HibpDownloaderException
from hibp_downloader.lib.logger import logger_get

logger = logger_get(__logger_name__)


async def httpx_debug_request(request):
    logger.debug(f"request: {request.method} {request.url}")
    if "_content" in vars(request) and request._content:  # noqa
        logger.debug(f"request-data: {request._content}")  # noqa


async def httpx_debug_response(response):
    logger.debug(
        f"response: {response.request.method} {response.request.url} "
        f"{response.status_code=} {response.http_version=} {response.headers=}"
    )


async def httpx_binary_response(url, etag=None, method="GET", encoding="gzip", debug=False):
    event_hooks = {}
    if debug:
        event_hooks["request"] = [httpx_debug_request]
        event_hooks["response"] = [httpx_debug_response]

    headers = {"User-Agent": f"{__title__}/{__version__}"}

    if encoding:
        headers["Accept-Encoding"] = encoding
    else:
        headers["Accept-Encoding"] = "identity"

    if etag:
        headers["If-None-Match"] = etag

    httpx_client = {
        "headers": headers,
        "http2": True,
        "timeout": 5,
        "follow_redirects": False,
        "trust_env": False,
    }

    if event_hooks:
        httpx_client["event_hooks"] = event_hooks

    async with httpx.AsyncClient(**httpx_client) as client:
        request = client.build_request(method=method, url=url)
        try:
            response = await client.send(request=request, stream=True)
        except (httpx.ConnectError, httpx.RemoteProtocolError):
            raise HibpDownloaderException(f"Unable to establish connection {url}")

        response.binary = b"".join([part async for part in response.aiter_raw()])

    return response



#
# Copyright [2023] Threat Patrols Pty Ltd (https://www.threatpatrols.com)
#

from . import __logger_name__, logger_get

logger = logger_get(__logger_name__)


class HibpDownloaderBaseException(Exception):
    def __init__(self, *args, **kwargs):
        log_message = " ".join([str(x) for x in args]).strip()
        if log_message:
            logger.error(f"{log_message}")
        if "detail" in kwargs:
            logger.error(f"{kwargs['detail']}".strip())
        super().__init__(*args)


class HibpDownloaderException(HibpDownloaderBaseException):
    pass



from dataclasses import asdict, dataclass, field
from datetime import datetime
from email.utils import parsedate_to_datetime
from enum import Enum
from typing import Union

from .hash_type import HashType


class PrefixMetadataDataSource(str, Enum):
    unknown_source_status = "unknown_source_status"
    local_source_ttl_cache = "local_source_ttl_cache"
    local_source_etag_match = "local_source_etag_match"
    remote_source_remote_cache = "remote_source_remote_cache"
    remote_source_origin_source = "remote_source_origin_source"


@dataclass()
class PrefixMetadata:
    prefix: str
    start_timestamp: Union[datetime, None] = field(default=None)
    hash_type: Union[HashType, None] = field(default=None)
    etag: Union[str, None] = field(default=None)
    bytes: Union[int, None] = field(default=0)
    server_timestamp: Union[str, datetime, None] = field(default=None)
    last_modified: Union[str, datetime, None] = field(default=None)
    content_encoding: Union[str, None] = field(default=None)
    content_checksum: Union[str, None] = field(default=None)
    data_source: Union[PrefixMetadataDataSource, None] = field(default=None)

    def __post_init__(self):
        if isinstance(self.server_timestamp, str):
            if ", " in self.server_timestamp:
                self.server_timestamp = parsedate_to_datetime(self.server_timestamp)
            else:
                self.server_timestamp = datetime.fromisoformat(self.server_timestamp)

        if isinstance(self.last_modified, str):
            if "," in self.last_modified:
                self.last_modified = parsedate_to_datetime(self.last_modified)
            else:
                self.last_modified = datetime.fromisoformat(self.last_modified)

    def as_dict(self):
        return {k: v for k, v in asdict(self).items() if v is not None}



from itertools import islice


def hex_sequence(hex_first: str = "00000", hex_last: str = "fffff", __string_length=5):
    if hex_first.startswith("0x"):
        hex_first = hex_first[2:]

    if hex_last.startswith("0x"):
        hex_last = hex_last[2:]

    hex_first = hex_first.lower()
    hex_last = hex_last.lower()

    value = int(hex_first, 16)
    while value <= int(str(hex_last), 16):
        yield str(hex(value))[2:].zfill(__string_length)
        value += 1


def iterable_chunker(iterable, size=10):
    iterable = iter(iterable)
    while True:
        x = tuple(islice(iterable, size))
        if not x:
            return
        yield x



import importlib
import os
import sys

import typer
from typing_extensions import Annotated

from .. import __help_epilog_footer__, __logger_name__, __title__, __version__, app_context
from ..exceptions import HibpDownloaderException
from ..lib.logger import logger_get

logger = logger_get(name=__logger_name__)
app = typer.Typer(
    add_completion=app_context.add_completion,
    no_args_is_help=app_context.no_args_is_help,
    rich_markup_mode="rich",
    epilog=(
        "Environment variables prefixed with HIBPDL_ that match their command-line equivalent may be used; for "
        "example use [bold]HIBPDL_DATA_PATH[/bold] to set the --data-path command option."
        "\n\n---\n"
        f"{__help_epilog_footer__}"
    ),
)


def invoke_app():
    try:
        load_commands()
        app()
    except HibpDownloaderException:
        exit(1)


@app.callback()
def main(
    ctx: typer.Context,
    data_path: Annotated[
        str,
        typer.Option(
            help="Path where a local-copy of the per-prefix HIBP pwned password data fragments are saved.",
            envvar="HIBPDL_DATA_PATH",
            show_envvar=False,
        ),
    ],
    metadata_path: Annotated[
        str,
        typer.Option(
            help="Path where metadata is saved; by default both data and metadata are saved in the same --data-path",
            envvar="HIBPDL_METADATA_PATH",
            show_envvar=False,
        ),
    ] = "",
    profiler: Annotated[
        bool,
        typer.Option(
            hidden=True, help="Enable the application performance profiler", envvar="HIBPDL_PROFILER", show_envvar=False
        ),
    ] = False,
    debug: Annotated[
        bool, typer.Option(help="Set logging to debug-level messages", envvar="HIBPDL_DEBUG", show_envvar=False)
    ] = False,
    quiet: Annotated[
        bool,
        typer.Option(
            help="Set logging to fatal-level messages; overrides --debug option",
            envvar="HIBPDL_QUIET",
            show_envvar=False,
        ),
    ] = False,
):
    """
    [bold]hibp-downloader[/bold] - Efficiently download new pwned password hashes from api.pwnedpasswords.com fast.
    """

    # return early if --help is all we need
    if "--help" in sys.argv:
        return

    # debug is captured at __init__; referenced here for happy linters
    if debug:
        ...

    # quiet is captured at __init__; referenced here for happy linters
    if quiet:
        ...

    # profiler is captured at __init__; referenced here for happy linters
    if profiler:
        ...

    # command context
    app_context.command = ctx.invoked_subcommand
    app_context.data_path = data_path
    app_context.metadata_path = metadata_path if metadata_path else data_path

    # start
    logger.info(f"{__title__}: v{__version__}")


def load_commands():
    loader_paths = [os.path.join(os.path.dirname(__file__), "..", "commands")]

    for loader_path in loader_paths:
        loader_path = os.path.realpath(loader_path)
        logger.debug(f"Loading command modules from {loader_path!r}")

        if loader_path not in sys.path:
            sys.path.append(loader_path)

        for filename in [f for f in os.listdir(loader_path) if os.path.isfile(os.path.join(loader_path, f))]:
            if filename.startswith("_"):
                continue
            module_name = filename.split(".")[0]

            try:
                module = importlib.import_module(module_name)
            except Exception as e:
                raise HibpDownloaderException(f"Failed importing command-module {module_name}", detail=e)

            logger.debug(f"Loaded command module {module_name!r} from {os.path.realpath(module.__file__)!r}")
            if loader_path not in module.__file__:
                logger.warning(
                    f"Command module {module_name!r} NOT loaded from {loader_path!r}; "
                    f"check for command-module filename collisions and consider renaming files."
                )

            for registered_group in app.registered_groups:
                if registered_group.name == module.command_name:
                    raise HibpDownloaderException(
                        f"Command {module.command_name!r} from {os.path.realpath(module.__file__)!r}"
                        "overrides already loaded command; consider changing the 'command_name' value."
                    )

            app.add_typer(module.command, name=module.command_name, rich_help_panel=module.command_section)



from .app_context import AppContext
from .hash_type import HashType
from .prefix_metadata import PrefixMetadata, PrefixMetadataDataSource
from .stats import QueueItemStats, QueueItemStatsCompute, QueueRunningStats



import gzip
import json
import os
from datetime import datetime
from typing import Union

import aiofiles
import aiofiles.os

from hibp_downloader.exceptions import HibpDownloaderException
from hibp_downloader.models import PrefixMetadata


async def append_stringfile(filename: str, content: str) -> None:
    if not await aiofiles.os.path.isfile(filename):
        return await save_bytesfile(os.path.dirname(filename), os.path.basename(filename), content.encode("utf8"))
    async with aiofiles.open(filename, mode="a") as f:
        await f.write(content)


async def save_bytesfile(pathname: str, filename: str, content: bytes, timestamp: Union[datetime, None] = None) -> None:
    await aiofiles.os.makedirs(pathname, exist_ok=True)
    async with aiofiles.open(os.path.join(pathname, filename), mode="wb") as f:
        await f.write(content)

    if timestamp:
        os.utime(os.path.join(pathname, filename), times=(timestamp.timestamp(), timestamp.timestamp()))


async def save_datafile(
    data_path: str, prefix: str, content: bytes, filename_suffix: str, timestamp: Union[datetime, None] = None
) -> None:
    await save_bytesfile(
        pathname=os.path.join(data_path, prefix[0:2], prefix[2:4]),
        filename=f"{prefix}.{filename_suffix}",
        content=content,
        timestamp=timestamp,
    )


async def save_metadatafile(metadata_path: str, prefix: str, metadata: PrefixMetadata) -> PrefixMetadata:
    def json_serial(data):
        if isinstance(data, datetime):
            return data.isoformat()
        raise HibpDownloaderException(f"Type {type(data)!r} not serializable")

    await save_bytesfile(
        pathname=os.path.join(metadata_path, prefix[0:2], prefix[2:4]),
        filename=f"{prefix}.meta",
        content=json.dumps(metadata.as_dict(), default=json_serial, separators=(",", ":")).encode("utf8"),
    )

    return metadata


async def load_bytesfile(pathname: str, filename: str) -> bytes:
    filename = os.path.join(pathname, filename)
    if not await aiofiles.os.path.isfile(filename):
        raise HibpDownloaderException(f"File not found {filename}")
    async with aiofiles.open(filename, mode="rb") as f:
        return await f.read()


async def load_datafile(
    data_path: str, prefix: str, filename_suffix: str, decompression_type=None, prepend_prefix=False
) -> str:
    data: bytes = await load_bytesfile(
        pathname=os.path.join(data_path, prefix[0:2], prefix[2:4]),
        filename=f"{prefix}.{filename_suffix}",
    )
    if decompression_type is None:
        pass
    elif decompression_type in ("gz", "gzip"):
        data = gzip.decompress(data)
    else:
        raise HibpDownloaderException(f"Unsupported decompression_type {decompression_type}")

    if prepend_prefix is False:
        return data.decode("utf8")

    data_lines = [f"{prefix.upper()}{x.upper()}" for x in data.decode("utf8").replace("\r", "").split("\n")]
    return "\n".join(data_lines)


async def load_metadata(metadata_path: str, prefix: str) -> PrefixMetadata:
    filename = os.path.join(metadata_path, prefix[0:2], prefix[2:4], f"{prefix}.meta")

    content = None
    if await aiofiles.os.path.isfile(filename):
        async with aiofiles.open(filename, "r") as f:
            content = await f.read()

    if not content:
        return PrefixMetadata(prefix=prefix)

    try:
        prefix_metadata = PrefixMetadata(**json.loads(content))
    except ValueError:
        return PrefixMetadata(prefix=prefix)

    if isinstance(prefix_metadata.server_timestamp, str):
        prefix_metadata.server_timestamp = datetime.fromisoformat(prefix_metadata.server_timestamp)
    if isinstance(prefix_metadata.last_modified, str):
        prefix_metadata.last_modified = datetime.fromisoformat(prefix_metadata.last_modified)

    return prefix_metadata



import asyncio
import hashlib
import os
from datetime import datetime
from itertools import repeat
from multiprocessing import Pool, Process, Queue

import typer
from typing_extensions import Annotated

from hibp_downloader import (
    __approx_gzip_bytes_per_hash__,
    __encoding_type__,
    __help_epilog_footer__,
    __local_cache_ttl_default__,
    __logger_name__,
    __logging_info_event_modulus__,
    __multiprocessing_prefixes_chunk_size__,
    __multiprocessing_processes_default__,
    __pwnedpasswords_api_url__,
    app_context,
)
from hibp_downloader.exceptions import HibpDownloaderException
from hibp_downloader.lib.filedata import load_metadata, save_datafile, save_metadatafile
from hibp_downloader.lib.generators import hex_sequence, iterable_chunker
from hibp_downloader.lib.http import httpx_binary_response
from hibp_downloader.lib.logger import logger_get
from hibp_downloader.models import (
    HashType,
    PrefixMetadata,
    PrefixMetadataDataSource,
    QueueItemStatsCompute,
    QueueRunningStats,
)

logger = logger_get(name=__logger_name__)

command = typer.Typer(no_args_is_help=False, epilog=__help_epilog_footer__)
command_name = "download"
command_section = "Commands"


results_queue = Queue()
results_queue_exit_sentinel = "__results_queue_exit_sentinel__"


@command.callback(invoke_without_command=True)
def main(
    hash_type: Annotated[
        HashType,
        typer.Option(help="Hash type to download from HIBP to the --data-path", case_sensitive=False),
    ] = HashType.sha1.value,
    first_hash: Annotated[
        str,
        typer.Option(help="Start the downloader from a specific hash prefix; trimmed to the first 5 characters"),
    ] = "00000",
    last_hash: Annotated[
        str,
        typer.Option(help="Stop the downloader at a hash prefix; trimmed to the first 5 characters"),
    ] = "fffff",
    processes: Annotated[
        int,
        typer.Option(
            help="Number of parallel processes to use; default value based on host CPU core count",
        ),
    ] = __multiprocessing_processes_default__,
    chunk_size: Annotated[
        int,
        typer.Option(
            help="Number of hash-prefixes to consume (asynchronously) per iteration per process",
        ),
    ] = __multiprocessing_prefixes_chunk_size__,
    force: Annotated[bool, typer.Option(help="Same as setting --local_cache_ttl=0 and --ignore-etag")] = False,
    ignore_etag: Annotated[
        bool, typer.Option(help="Do not use request etag headers to manage local/remote cached data")
    ] = False,
    local_cache_ttl: Annotated[
        int,
        typer.Option(
            help="Time-to-live (seconds) on local metadata cache items; "
            "prevents requesting the same data twice in this period"
        ),
    ] = __local_cache_ttl_default__,
):
    """
    Download new pwned password hash data from HIBP and update the local --data-path data storage path; use [bold cyan]download --help[/bold cyan] for more.
    """

    logger.debug(f"Starting command {app_context.command!r} from {os.path.basename(__file__)!r}")

    if not os.path.isdir(app_context.data_path):
        logger.warning(f"Data path {app_context.data_path!r} does not exist, creating it now...")
        os.makedirs(app_context.data_path, exist_ok=True)

    if not os.path.isdir(app_context.metadata_path):
        logger.warning(f"Metadata path {app_context.metadata_path!r} does not exist, creating it now...")
        os.makedirs(app_context.metadata_path, exist_ok=True)

    logger.info(f"data-path {app_context.data_path!r}")
    logger.info(f"metadata-path {app_context.metadata_path!r}")

    if force:
        ignore_etag = True
        local_cache_ttl = 0

    results_queue_process = Process(target=results_queue_processor, args=(results_queue,))
    results_queue_process.start()

    with Pool(
        processes=processes, initializer=results_queue_initialize, initargs=(results_queue,)
    ) as multiprocessing_pool:
        prefix_chunks = iterable_chunker(
            iterable=hex_sequence(hex_first=first_hash[0:5], hex_last=last_hash[0:5]), size=chunk_size
        )

        multiprocessing_pool.starmap(
            pwnedpasswords_get_and_store_asyncio_run,
            zip(
                prefix_chunks,
                repeat(hash_type),
                repeat(os.path.join(app_context.data_path, hash_type.value)),  # data_path
                repeat(os.path.join(app_context.metadata_path, hash_type.value)),  # metadata_path
                repeat(__encoding_type__),  # encoding_type: read comments in __init__ before wanting Brotli
                repeat(ignore_etag),
                repeat(local_cache_ttl),
                repeat(logger),
            ),
        )

        multiprocessing_pool.close()
        multiprocessing_pool.join()
        results_queue.put(results_queue_exit_sentinel)

    results_queue_process.join()


def pwnedpasswords_get_and_store_asyncio_run(*args):
    asyncio.run(pwnedpasswords_get_and_store_gather(*args))


async def pwnedpasswords_get_and_store_gather(*args) -> None:
    """
    Gather the async tasks to get pwnedpasswords per prefix-chunk item
     - First arg is the list of prefixes in this chunk
     - Last arg is the results_queue
    """
    prefixes = args[0]
    metadata_results = await asyncio.gather(
        *[pwnedpasswords_get_and_store_async(prefix, *args[1:-1]) for prefix in prefixes],
    )
    results_queue.put(metadata_results)


async def pwnedpasswords_get_and_store_async(
    prefix: str,
    hash_type: HashType,
    data_path: str,
    metadata_path: str,
    encoding_type: str,
    ignore_etag: bool,
    local_cache_ttl: int,
    logger_: logger = None,
):
    start_timestamp = datetime.now().astimezone()

    if not logger_:
        logger_ = logger_get(name=None)

    logger_.debug(
        f"{prefix=} hash_type='{hash_type.value}' {encoding_type=} {ignore_etag=} {local_cache_ttl=} {start_timestamp=}"
    )

    # get existing metadata if available
    metadata = await load_metadata(metadata_path=metadata_path, prefix=prefix)

    etag = None
    if metadata.data_source:
        if metadata.server_timestamp:
            local_ttl = local_cache_ttl - (datetime.now().astimezone() - metadata.server_timestamp).seconds
            if local_ttl > 0:
                logger_.debug(f"Skipping {prefix}; local-cache has {local_ttl} time-to-live")
                return PrefixMetadata(
                    prefix=prefix,
                    data_source=PrefixMetadataDataSource.local_source_ttl_cache,
                    start_timestamp=start_timestamp,
                )
        if not ignore_etag:
            etag = metadata.etag

    # download with etag setting
    binary, metadata = await pwnedpasswords_get(prefix, hash_type=hash_type, etag=etag, encoding=encoding_type)
    metadata.start_timestamp = start_timestamp

    # filename suffix based on encoding
    if encoding_type is None or encoding_type.lower() == "identity":
        filename_suffix = "txt"
    elif encoding_type.lower() == "gzip" or encoding_type.lower() == "gz":
        filename_suffix = "gz"  # extension makes shell command completion for zcat, zgrep and others work nicely
    else:
        filename_suffix = encoding_type.lower()

    # save
    if binary:
        await save_datafile(
            data_path=data_path,
            prefix=prefix,
            content=binary,
            filename_suffix=filename_suffix,
            timestamp=metadata.last_modified,
        )

    if metadata.data_source not in (
        PrefixMetadataDataSource.local_source_ttl_cache,
        PrefixMetadataDataSource.unknown_source_status,
    ):
        await save_metadatafile(metadata_path=metadata_path, prefix=prefix, metadata=metadata)

    return metadata


async def pwnedpasswords_get(prefix: str, hash_type: HashType, etag: str, encoding: str, httpx_debug: bool = False):
    url = f"{__pwnedpasswords_api_url__}/range/{prefix}"
    if hash_type == HashType.ntlm:
        url += "?mode=ntlm"

    try:
        response = await httpx_binary_response(url=url, etag=etag, encoding=encoding, debug=httpx_debug)
    except HibpDownloaderException:
        return None, PrefixMetadata(prefix=prefix, data_source=PrefixMetadataDataSource.unknown_source_status)

    metadata = PrefixMetadata(
        prefix=prefix,
        hash_type=hash_type,
        etag=response.headers.get("etag"),
        bytes=len(response.binary),
        server_timestamp=response.headers.get("date"),
        last_modified=response.headers.get("last-modified"),
        content_encoding=response.headers.get("content-encoding"),
        content_checksum=hashlib.sha256(response.binary).hexdigest() if response.binary else None,
    )

    if response.status_code == 304:  # HTTP 304 Not Modified status
        metadata.data_source = PrefixMetadataDataSource.local_source_etag_match
    elif response.status_code == 200:
        if response.headers.get("cf-cache-status").upper() == "HIT":
            metadata.data_source = PrefixMetadataDataSource.remote_source_remote_cache
        else:
            metadata.data_source = PrefixMetadataDataSource.remote_source_origin_source
    else:
        metadata.data_source = PrefixMetadataDataSource.unknown_source_status

    return response.binary, metadata


def results_queue_initialize(q):
    global results_queue
    results_queue = q


def results_queue_processor(q: Queue):
    running_stats = QueueRunningStats()

    while True:
        metadata_items = q.get()
        if metadata_items == results_queue_exit_sentinel:
            running_stats.end_trigger()
            logger.info(f"Finished in {round(running_stats.run_time/60,1)}min")
            break

        if metadata_items:
            running_stats.add_item_stats(item=QueueItemStatsCompute(metadata_items).stats)

        if running_stats.queue_item_count % __logging_info_event_modulus__ == 0:
            logger.info(
                f"prefix={running_stats.prefix_latest} "
                f"source=[lc:{running_stats.local_source_ttl_cache_count_sum} "
                f"et:{running_stats.local_source_etag_match_count_sum} "
                f"rc:{running_stats.remote_source_remote_cache_count_sum} "
                f"ro:{running_stats.remote_source_origin_source_count_sum} "
                f"xx:{running_stats.unknown_source_status_count_sum}] "
                f"runtime_rate=[{to_mbytes(running_stats.bytes_received_rate_total * 8, 1)}MBit/s "
                f"{int(running_stats.request_rate_total)}req/s "
                f"~{int(running_stats.bytes_received_rate_total / __approx_gzip_bytes_per_hash__)}H/s] "
                f"runtime={round(running_stats.run_time/60,1)}min "
                f"download={to_mbytes(running_stats.bytes_received_sum, 1)}MB"
            )


def to_mbytes(value, rounding=None):
    if value is not None:
        return round(value / 1024 / 1024, rounding) if rounding else value / 1024 / 1024



import hashlib

from .md4 import MD4


def hashed_sha1(value: str):
    return hashlib.sha1(value.encode(), usedforsecurity=False).hexdigest().lower()


def hashed_ntlm(value: str):
    return MD4(value.encode("utf-16le")).hexdigest().lower()
